from PyQt6 import QtWidgets

from vars_gridview.ui.settings.tabs.AbstractSettingsTab import AbstractSettingsTab


class SQLTab(AbstractSettingsTab):
    """
    SQL server tab.
    """
    
    def __init__(self, parent=None):
        super().__init__('SQL', parent=parent)
        
        self.sql_url_edit = QtWidgets.QLineEdit(self._settings.sql_url.value)
        self.sql_url_edit.textChanged.connect(self.settingsChanged.emit)
        
        self.user_edit = QtWidgets.QLineEdit(self._settings.sql_user.value)
        self.user_edit.textChanged.connect(self.settingsChanged.emit)
        
        self.password_edit = QtWidgets.QLineEdit(self._settings.sql_password.value)
        self.password_edit.setEchoMode(QtWidgets.QLineEdit.EchoMode.Password)
        self.password_edit.textChanged.connect(self.settingsChanged.emit)
        
        self.database_edit = QtWidgets.QLineEdit(self._settings.sql_database.value)
        self.database_edit.textChanged.connect(self.settingsChanged.emit)
        
        self.arrange()
    
    def arrange(self):
        layout = QtWidgets.QFormLayout()
        
        layout.addRow('Server URL', self.sql_url_edit)
        layout.addRow('Username', self.user_edit)
        layout.addRow('Password', self.password_edit)
        layout.addRow('', QtWidgets.QLabel('The password is stored in plain text. For this reason, it is recommended to use a read-only login.'))
        layout.addRow('Database', self.database_edit)
        
        self.setLayout(layout)
    
    def apply_settings(self):
        self._settings.sql_url.value = self.sql_url_edit.text()
        self._settings.sql_user.value = self.user_edit.text()
        self._settings.sql_password.value = self.password_edit.text()
        self._settings.sql_database.value = self.database_edit.text()



from PyQt6 import QtWidgets

from vars_gridview.lib.settings import SettingsManager


class LoginDialog(QtWidgets.QDialog):
    """
    Dialog to get a username and password. Completer optional for username.
    """
    
    class LoginForm(QtWidgets.QWidget):
        """
        Login form widget.
        """
        
        def __init__(self, parent=None, completer=None):
            super().__init__(parent)
            
            settings = SettingsManager.get_instance()
            
            self._username_line_edit = QtWidgets.QLineEdit()
            if completer is not None:
                self._username_line_edit.setCompleter(completer)
            
            self._password_line_edit = QtWidgets.QLineEdit()
            self._password_line_edit.setEchoMode(QtWidgets.QLineEdit.EchoMode.Password)
            
            raziel_url = settings.raz_url.value
            self._raziel_url_line_edit = QtWidgets.QLineEdit()
            self._raziel_url_line_edit.setText(raziel_url)
            self._raziel_url_line_edit.setPlaceholderText(raziel_url)
            
            self._arrange()
        
        def _arrange(self):
            layout = QtWidgets.QFormLayout()
        
            layout.addRow('Username:', self._username_line_edit)
            layout.addRow('Password:', self._password_line_edit)
            layout.addRow('Config server:', self._raziel_url_line_edit)
            
            self.setLayout(layout)
        
        @property
        def credentials(self):
            return self._username_line_edit.text(), self._password_line_edit.text()
        
        @property
        def raziel_url(self):
            return self._raziel_url_line_edit.text()
    
    def __init__(self, parent=None, completer=None):
        super().__init__(parent)
        
        self.setWindowTitle('Login')
        
        self._login_form = LoginDialog.LoginForm(self, completer)
        
        self._dialog_buttons = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.StandardButton.Ok | 
            QtWidgets.QDialogButtonBox.StandardButton.Cancel
        )
        self._dialog_buttons.accepted.connect(self.accept)
        self._dialog_buttons.rejected.connect(self.reject)
        self._dialog_buttons.button(QtWidgets.QDialogButtonBox.StandardButton.Ok).setText('Login')
        
        self._arrange()
    
    def _arrange(self):
        layout = QtWidgets.QVBoxLayout()
        
        layout.addWidget(self._login_form)
        layout.addWidget(self._dialog_buttons)
        
        self.setLayout(layout)
    
    @property
    def credentials(self):
        return self._login_form.credentials
    
    @property
    def raziel_url(self):
        return self._login_form.raziel_url



"""
M3 REST API clients.
"""

import requests
import requests.auth


class JWTAuth(requests.auth.AuthBase):
    """
    JWT Auth class for requests.
    """

    def __init__(self, token: str):
        self._token = token

    def __call__(self, r):
        r.headers['Authorization'] = 'BEARER {}'.format(self._token)
        return r


class NotAuthenticated(Exception):
    """
    Raised when a client is not authenticated.
    """
    pass


def needs_auth(f):
    """
    Decorator to ensure that the client is authenticated.
    Raises NotAuthenticated if not.
    """
    def wrapper(self, *args, **kwargs):
        if not self.authenticated:
            raise NotAuthenticated
        else:
            return f(self, *args, **kwargs)
    return wrapper

class M3Client:
    """
    M3 microservice client.
    """
    
    def __init__(self, base_url: str, api_key: str = None):
        self._session = requests.Session()
        
        self.base_url = base_url
        
        if api_key is not None:
            self.authenticate(api_key)
    
    @property
    def base_url(self) -> str:
        return self._base_url

    @base_url.setter
    def base_url(self, base_url: str):
        self._base_url = base_url.rstrip('/')

    def url_to(self, path: str) -> str:
        """
        Get the full URL for a path.
        """
        return self._base_url + path

    def get(self, path: str, **kwargs) -> requests.Response:
        return self._session.get(self.url_to(path), **kwargs)
    
    def put(self, path: str, **kwargs) -> requests.Response:
        return self._session.put(self.url_to(path), **kwargs)
    
    def post(self, path: str, **kwargs) -> requests.Response:
        return self._session.post(self.url_to(path), **kwargs)
    
    def delete(self, path: str, **kwargs) -> requests.Response:
        return self._session.delete(self.url_to(path), **kwargs)

    @property
    def authenticated(self) -> bool:
        """
        True if the client is authenticated.
        """
        return self._session.auth is not None

    def authenticate(self, api_key: str, auth_path: str = '/auth'):
        """
        Authenticate the client with the provided API key.
        """
        self._session.auth = None
        
        response = self.post(auth_path, headers={'Authorization': f'APIKEY {api_key}'})
        response.raise_for_status()
        
        token = response.json()['access_token']
        self._session.auth = JWTAuth(token)


class AnnosaurusClient(M3Client):
    """
    Annosaurus (v1) client.
    """
    
    def url_to(self, path: str) -> str:
        return super().url_to(path)
    
    @needs_auth
    def update_association(self, association_uuid: str, data: dict) -> requests.Response:
        return self.put(f'/associations/{association_uuid}', data=data)

    @needs_auth
    def delete_association(self, association_uuid: str) -> requests.Response:
        return self.delete(f'/associations/{association_uuid}')
    
    @needs_auth
    def update_observation(self, observation_uuid: str, data: dict) -> requests.Response:
        return self.put(f'/observations/{observation_uuid}', data=data)


class VampireSquidClient(M3Client):
    """
    Vampire Squid (v1) client.
    """
    
    def url_to(self, path: str) -> str:
        return super().url_to('/v1' + path)
    
    def get_videos_at_timestamp(self, timestamp: str) -> requests.Response:
        return self.get(f'/videos/timestamp/{timestamp}')


class VARSUserServerClient(M3Client):
    """
    VARS user server (v1) client.
    """

    def url_to(self, path: str) -> str:
        return super().url_to(path)
    
    def get_all_users(self) -> requests.Response:
        return self.get('/users')


class VARSKBServerClient(M3Client):
    """
    VARS KB server (v1) client.
    """
    
    def url_to(self, path: str) -> str:
        return super().url_to(path)
    
    def get_concepts(self) -> requests.Response:
        return self.get('/concept')
    
    def get_parts(self) -> requests.Response:
        return self.get('/phylogeny/taxa/organism part')

    def get_phylogeny_taxa(self, concept: str) -> requests.Response:
        return self.get(f'/phylogeny/taxa/{concept}')



# -*- coding: utf-8 -*-
"""
rectlabel.py -- Tools to implement a labeling UI for bounding boxes in images
Copyright 2020  Monterey Bay Aquarium Research Institute
Distributed under MIT license. See license.txt for more information.

"""
import json
from typing import Optional, Union

import numpy as np

from vars_gridview.lib.m3.operations import delete_bounding_box, update_bounding_box_data, update_bounding_box_part, update_observation_concept


class VARSLocalization:
    """Representation of VARS localizations (bounding boxes)"""

    def __init__(
        self, x: int, y: int, width: int, height: int, image_reference_uuid: Optional[str] = None, **meta
    ):
        self._x = x
        self._y = y
        self._width = width
        self._height = height
        self.image_reference_uuid = image_reference_uuid
        self.observation_uuid = None
        self.association_uuid = None
        self.imaged_moment_uuid = None
        self.meta = meta
        self._concept = None
        self._part = None

        # Dirty flags, controls what gets updated in VARS when pushed
        self._dirty_concept = False
        self._dirty_part = False
        self._dirty_box = False
        self._dirty_verifier = False

        # Deleted flag
        self._deleted = False

        # Dreaded back-reference
        self.rect = None

    @staticmethod
    def from_json(data: Union[str, dict]):
        if type(data) == str:
            data = json.loads(data)

        return VARSLocalization(**data)

    @property
    def json(self):
        return {
            'x': self._x,
            'y': self._y,
            'width': self._width,
            'height': self._height,
            'image_reference_uuid': self.image_reference_uuid,
            **self.meta,
        }

    @property
    def json_str(self):
        return json.dumps(self.json)

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def concept(self):
        return self._concept

    @property
    def part(self):
        return self._part

    @property
    def text_label(self):
        if self._part == 'self':
            return self._concept

        return '{} {}'.format(self._concept, self._part)

    @property
    def xf(self):
        return self._x + self._width

    @property
    def yf(self):
        return self._y + self._height

    @property
    def box(self):
        return self._x, self._y, self.xf, self.yf

    @property
    def verified(self):
        return 'verifier' in self.meta

    @property
    def deleted(self):
        return self._deleted

    def set_box(self, x: int, y: int, width: int, height: int):
        if self.x != x or self.y != y or self.width != width or self.height != height:
            self._dirty_box = True

        self._x = int(x)
        self._y = int(y)
        self._width = int(width)
        self._height = int(height)

    def set_concept(self, concept: str, part: str):
        if self._concept is not None and concept != self._concept:
            self._dirty_concept = True

        if self._part is not None and part != self._part:
            self._dirty_part = True

        self._concept = concept
        self._part = part

    def set_verified_concept(self, concept, part, verifier):
        self.set_concept(concept, part)
        self.meta['verifier'] = verifier
        self._dirty_verifier = True

    def get_roi(self, image: np.ndarray):
        return image[self._y : self.yf, self._x : self.xf]

    @property
    def valid_box(self):
        return self.xf > self.x and self.yf > self.y

    def in_bounds(self, min_x, min_y, max_x, max_y):
        return (
            self.x >= min_x
            and self.y >= min_y
            and self.xf <= max_x
            and self.yf <= max_y
        )

    def delete(self):
        delete_bounding_box(self.association_uuid)
        self._deleted = True

    def push_changes(self, verifier: str):
        do_modify_box = False

        if self._dirty_concept:
            update_observation_concept(self.observation_uuid, self._concept, verifier)
            self._dirty_concept = False
            do_modify_box = True

        if self._dirty_part:
            update_bounding_box_part(self.association_uuid, self._part)
            self._dirty_part = False
            do_modify_box = True

        if self._dirty_box:
            self.meta['generator'] = 'gridview'  # Only changes when box moved/resized
            self.meta['observer'] = verifier
            self._dirty_box = False
            do_modify_box = True

        if self._dirty_verifier:
            do_modify_box = True
            self._dirty_verifier = False

        if do_modify_box:
            update_bounding_box_data(self.association_uuid, self.json)



import sys
from typing import Tuple

import pymssql

from vars_gridview.lib.settings import SettingsManager
from vars_gridview.lib.constants import BASE_QUERY_FILE


SQL_CONNECTION = None
BASE_QUERY = None


class ORList:
    @staticmethod
    def typestr(value):
        if isinstance(value, str):
            return '%s'
        elif isinstance(value, int):
            return '%d'
        else:
            raise ValueError('Unsupported type: {}'.format(type(value)))

    def __init__(self, key: str, values=None):
        self._key = key
        self._values = values or []

    def __iadd__(self, value):
        self._values.append(value)
        return self

    @property
    def form(self):
        return (
            '('
            + ' OR '.join(
                [
                    '{} = {}'.format(self._key, ORList.typestr(value))
                    for value in self._values
                ]
            )
            + ')'
        )

    @property
    def values(self):
        return self._values


class ConstraintSpec:
    def __init__(self, lists=None):
        self._lists = lists or []

    @property
    def form(self):
        if len(self._lists) == 0:
            return '1=1'  # Accept anything
        return ' AND '.join(
            [list.form for list in self._lists]
        )  # AND together all forms

    @property
    def values(self):
        return [value for l in self._lists for value in l.values]  # Flatten all values

    @classmethod
    def from_dict(cls, d):
        return cls(lists=[ORList(key, values) for key, values in d.items()])


def get_base_query() -> str:
    """
    Read the base query.
    """
    global BASE_QUERY
    if not BASE_QUERY:
        with open(BASE_QUERY_FILE) as f:
            BASE_QUERY = f.read()
    
    return BASE_QUERY


def connect(server_url: str, user: str, password: str, database: str):
    """
    Initialize the connection to the SQL server.
    """
    global SQL_CONNECTION
    try:
        SQL_CONNECTION = pymssql.connect(server=server_url, user=user, password=password, database=database)
    except pymssql.DatabaseError as e:
        pass  # TODO: log an error
        print(f'Failed to connect to SQL server: {e}')
        sys.exit(1)


def connect_from_settings():
    """
    Connect.
    """
    settings = SettingsManager.get_instance()
    
    connect(settings.sql_url.value, settings.sql_user.value, settings.sql_password.value, settings.sql_database.value)


def query(constraint_dict: dict) -> Tuple[list, list]:
    if not SQL_CONNECTION:
        raise Exception('No connection to SQL server')
    
    constraint_spec = ConstraintSpec.from_dict(constraint_dict)

    cursor = SQL_CONNECTION.cursor()
    cursor.execute(
        get_base_query().format(filters=constraint_spec.form), tuple(constraint_spec.values)
    )

    return cursor.fetchall(), [
        i[0] for i in cursor.description
    ]  # Data and column names



# -*- coding: utf-8 -*-
"""
gridview.py -- A small python app to display ROIs and edit labels driven by FathomNet annotations
Copyright 2020  Monterey Bay Aquarium Research Institute
Distributed under MIT license. See license.txt for more information.

The app loads images from a directory and then sorts them by label, time, or height. Based on the max
image size and the scree width, it divides the layout into columns and rows and then
fills these with RectWidgets. Each rect widget controls an ROI and annotation and
new annotations can be applied to groups of selected widgets

The app also provides a view similar to rectlabel, where existing annotations can be dragged to adjust and
new annotations can be added.

This was inspired from a PyQt4 GraphicsGridLayout example found here:
http://synapses.awardspace.info/pages-scripts/python/pages/python-pyqt_qgraphicsview-thumbnails-grid.py.html
"""

import datetime
import json
import os
import sys
import webbrowser
from typing import Optional, Tuple
from pathlib import Path

import cv2
import pyqtgraph as pg
import qdarkstyle
from PyQt6.QtCore import QFile, QFileInfo, QTextStream, QSettings, pyqtSlot
from PyQt6.QtGui import QAction
from PyQt6.QtWidgets import (
    QApplication,
    QCompleter,
    QMainWindow,
    QMessageBox,
)

from vars_gridview.lib import m3, sql, constants, raziel
from vars_gridview.lib.log import LOGGER
from vars_gridview.lib.boxes import BoxHandler
from vars_gridview.lib.image_mosaic import ImageMosaic
from vars_gridview.lib.m3.operations import get_kb_concepts, get_kb_parts, get_vars_imaged_moment, get_videos_at_datetime
from vars_gridview.lib.settings import SettingsManager
from vars_gridview.lib.sort_methods import AreaSort, HeightSort, ImageReferenceUUIDSort, LabelSort, MeanHueSort, MeanIntensitySort, WidthSort, RegionMeanHueSort, DepthSort
from vars_gridview.lib.widgets import RectWidget
from vars_gridview.ui.LoginDialog import LoginDialog
from vars_gridview.ui.QueryDialog import QueryDialog
from vars_gridview.ui.settings.SettingsDialog import SettingsDialog

# Define main window class from template
CWD = Path(__file__).parent
ASSETS_DIR = CWD.parent / 'assets'
UI_FILE_PATH = ASSETS_DIR / 'gridview.ui'
WindowTemplate, TemplateBaseClass = pg.Qt.loadUiType(UI_FILE_PATH)

GUI_SETTINGS = QSettings(str(constants.GUI_SETTINGS_FILE), QSettings.Format.IniFormat)

ENABLED_SORT_METHODS = [
    ImageReferenceUUIDSort,
    LabelSort,
    WidthSort,
    HeightSort,
    AreaSort,
    MeanIntensitySort,
    MeanHueSort,
    RegionMeanHueSort,
    DepthSort,
]


class MainWindow(TemplateBaseClass):
    """
    Main application window.
    """
    
    def __init__(self, app):
        QMainWindow.__init__(self)
        TemplateBaseClass.__init__(self)
        
        self._app = app

        # Create the main window
        self.ui = WindowTemplate()
        self.ui.setupUi(self)

        # Set the window title
        self.setWindowTitle(constants.APP_NAME)

        # Restore and style GUI
        self._restore_gui()
        self._style_gui()
        
        self.verifier = None  # The username of the current verifier
        self.endpoints = None  # The list of endpoint data from Raziel

        self.last_selected_roi = None  # Last selected ROI

        self.image_mosaic = (
            None  # Image mosaic (holds the thumbnails as a grid of RectWidgets)
        )
        self.box_handler = None  # Box handler (handles the ROIs and annotations)
        
        self.cached_moment_concepts = {}  # Cache for imaged moment -> set of observed concepts

        # Connect signals to slots
        self.ui.discardButton.clicked.connect(self.move_to_discard)
        self.ui.clearSelections.clicked.connect(self.clear_selected)
        self.ui.labelSelectedButton.clicked.connect(self.update_labels)
        self.ui.zoomSpinBox.valueChanged.connect(self.update_zoom)
        self.ui.sortMethod.currentTextChanged.connect(self.update_layout)
        self.ui.hideLabeled.stateChanged.connect(self.update_layout)
        self.ui.styleComboBox.currentTextChanged.connect(self._style_gui)
        self.ui.openVideo.clicked.connect(self.open_video)
        
        self.settings_dialog = SettingsDialog(self)

        self._launch()

    @property
    def loaded(self):
        return self.image_mosaic is not None and self.box_handler is not None

    def _launch(self):
        """
        Perform additional on-launch setup after the window is shown.
        """
        LOGGER.info('Launching application')
        
        # Run the login procedure
        login_ok = self._login_procedure()
        if not login_ok:
            LOGGER.error('Login failed')
            QMessageBox.critical(self, "Login failed", "Login failed, exiting.")
            sys.exit(1)

        # Set up the label combo boxes
        self._setup_label_boxes()
        
        # Set up the sort method combo box
        self._setup_sort_methods()

        # Set up the menu bar
        self._setup_menu_bar()
        
        LOGGER.info('Launch successful')
    
    def _get_login(self) -> Optional[Tuple[str, str, str]]:
        """
        Prompt a login dialog and return the username, password, and Raziel URL. If failed, returns None.
        """
        login_dialog = LoginDialog(parent=self)
        ok = login_dialog.exec()
        
        if not ok:
            return None
        
        return (*login_dialog.credentials, login_dialog.raziel_url)
    
    def _login_procedure(self) -> bool:
        """
        Perform the full login + authentication procedure. Return True on success, False on fail.
        """
        login = self._get_login()
        if login is None:  # Login failed
            return False
        
        username, password, raziel_url = login  # unpack the credentials
        
        # Update the Raziel URL setting
        settings = SettingsManager.get_instance()
        if settings.raz_url.value != raziel_url:
            LOGGER.debug(f'Updating Raziel URL setting to {raziel_url}')
            settings.raz_url.value = raziel_url
        
        # Authenticate Raziel + get endpoint data
        endpoints = self._auth_raziel(raziel_url, username, password)
        if endpoints is None:  # Authentication failed
            return False
        
        # Authenticate M3 modules
        ok = self._setup_m3(endpoints)
        if not ok:
            return False
        
        # Connect to the database
        sql.connect_from_settings()
        
        # Set the verifier and endpoint data
        self.verifier = username
        self.endpoints = endpoints
        
        return True
    
    def _auth_raziel(self, raziel_url, username, password) -> Optional[list]:
        """
        Authenticate with Raziel. Return endpoints list on success, None on fail.
        """
        LOGGER.debug(f'Attempting to authenticate user {username} with Raziel at {raziel_url}')
        try:
            endpoints = raziel.authenticate(raziel_url, username, password)
            return endpoints
        except Exception as e:
            LOGGER.error(f'Raziel authentication failed: {e}')
            QMessageBox.critical(self, "Authentication failed", f"Failed to authenticate with the configuration server. Check your username and password.\n\n{e}")
    
    def _setup_m3(self, endpoints: list) -> bool:
        """
        Setup the M3 modules from a list of authenticated Raziel endpoint dicts. Return True on success, False on fail.
        """
        LOGGER.debug('Attempting to set up M3')
        try:
            m3.setup_from_endpoint_data(endpoints)
            return True
        except ValueError as e:
            LOGGER.error(f'M3 setup failed: {e}')
            QMessageBox.critical(self, "M3 setup failed", f"M3 setup failed: {e}")
            return False

    def _setup_menu_bar(self):
        """
        Populate the menu bar with menus and actions.
        """
        menu_bar = self.ui.menuBar
        
        file_menu = menu_bar.addMenu('&File')
        
        settings_action = QAction('&Settings', self)
        settings_action.setShortcut('Ctrl+,')
        settings_action.triggered.connect(self._open_settings)
        file_menu.addAction(settings_action)

        query_menu = menu_bar.addMenu('&Query')

        query_action = QAction('&Query', self)
        query_action.setShortcut('Ctrl+Q')
        query_action.triggered.connect(self._do_query)
        query_menu.addAction(query_action)
    
    def _open_settings(self):
        """
        Open the settings dialog.
        """
        self.settings_dialog.show()
    
    def _setup_from_settings(self):
        """
        Propagate the settings to the app.
        """
        m3.setup_from_settings()
        sql.connect_from_settings()

    def _do_query(self):
        """
        Perform a query based on the filter string.
        """
        # Show a query dialog
        constraint_dict = self.query()
        if constraint_dict is None:  # User cancelled, do nothing
            return
        else:  # Unload
            self.last_selected_roi = None
            self.image_mosaic = None
            self.box_handler = None

        # Run the query
        query_data, query_headers = sql.query(constraint_dict)

        # Grab the beholder info
        beholder_endpoint = next(e for e in self.endpoints if e['name'] == 'beholder')
        beholder_url = beholder_endpoint['url']
        beholder_api_key = beholder_endpoint['secret']

        # Create the image mosaic
        self.image_mosaic = ImageMosaic(
            self.ui.roiGraphicsView,
            query_data,
            query_headers,
            self.rect_click,
            self.verifier,
            beholder_url,
            beholder_api_key,
            zoom=self.ui.zoomSpinBox.value() / 100,
        )

        self.image_mosaic.hide_discarded = False
        self.image_mosaic.hide_to_review = False
        self.image_mosaic.hide_labeled = self.ui.hideLabeled.isChecked()

        # Render
        sort_method = self.ui.sortMethod.currentData()
        self.image_mosaic.render_mosaic(sort_method=sort_method)

        # Show some stats about the images and annotations
        self.statusBar().showMessage(
            'Loaded '
            + str(self.image_mosaic.n_images)
            + ' images and '
            + str(self.image_mosaic.n_localizations)
            + ' localizations.'
        )

        # Create the box handler
        self.box_handler = BoxHandler(
            self.ui.roiDetailGraphicsView,
            self.image_mosaic,
            all_labels=get_kb_concepts(),
            verifier=self.verifier,
        )

    def _setup_label_boxes(self):
        """
        Populate the label combo boxes
        """
        # Set up the combo boxes
        self.ui.labelComboBox.clear()
        concepts = [''] + sorted([c for c in get_kb_concepts() if c != ''])
        self.ui.labelComboBox.addItems(concepts)
        self.ui.labelComboBox.completer().setCompletionMode(QCompleter.CompletionMode.PopupCompletion)

        self.ui.partComboBox.clear()
        parts = [''] + sorted([p for p in get_kb_parts() if p != ''])
        self.ui.partComboBox.addItems(parts)
        self.ui.partComboBox.completer().setCompletionMode(QCompleter.CompletionMode.PopupCompletion)

    def _setup_sort_methods(self):
        """
        Populate the sort method combo box
        """
        self.ui.sortMethod.clear()
        for method in ENABLED_SORT_METHODS:
            self.ui.sortMethod.addItem(method.NAME, userData=method)

    def _restore_gui(self):
        """
        Restore window size and splitter states
        """
        finfo = QFileInfo(GUI_SETTINGS.fileName())
        if finfo.exists() and finfo.isFile():
            self.restoreGeometry(GUI_SETTINGS.value("geometry"))
            self.restoreState(GUI_SETTINGS.value("windowState"))
            self.ui.splitter1.restoreState(GUI_SETTINGS.value("splitter1state"))
            self.ui.splitter2.restoreState(GUI_SETTINGS.value("splitter2state"))

    def _save_gui(self):
        GUI_SETTINGS.setValue('geometry', self.saveGeometry())
        GUI_SETTINGS.setValue('windowState', self.saveState())
        GUI_SETTINGS.setValue('splitter1state', self.ui.splitter1.saveState())
        GUI_SETTINGS.setValue('splitter2state', self.ui.splitter2.saveState())

    def update_labels(self):
        concept = self.ui.labelComboBox.currentText()
        part = self.ui.partComboBox.currentText()

        if concept not in get_kb_concepts() and concept != '':
            QMessageBox.critical(
                self, 'Bad Concept', f'Bad concept \"{concept}\". Canceling.'
            )
            return
        if part not in get_kb_parts() and part != '':
            QMessageBox.critical(self, 'Bad Part', f'Bad part \"{part}\". Canceling.')
            return

        to_label = self.image_mosaic.get_selected()
        if len(to_label) > 1:
            opt = QMessageBox.question(
                self,
                'Confirm Label',
                'Label {} localizations?'.format(
                    len(to_label)
                ),
                defaultButton=QMessageBox.StandardButton.No,
            )
        else:
            opt = QMessageBox.StandardButton.Yes

        if opt == QMessageBox.StandardButton.Yes:
            # Apply labels to all selected localizations, push to VARS
            self.image_mosaic.apply_label(concept, part)

            # Update the label of the selected localization in the image view (if necessary)
            self.box_handler.update_labels()

    @pyqtSlot()
    def move_to_discard(self):
        if not self.loaded:
            return

        to_delete = self.image_mosaic.get_selected()
        opt = QMessageBox.question(
            self,
            'Confirm Deletion',
            'Delete {} localizations?\nThis operation cannot be undone.'.format(
                len(to_delete)
            ),
            defaultButton=QMessageBox.StandardButton.No,
        )
        if opt == QMessageBox.StandardButton.Yes:
            self.image_mosaic.delete_selected()

    @pyqtSlot()
    def clear_selected(self):
        if not self.loaded:
            return

        self.image_mosaic.clear_selected()

    @pyqtSlot()
    def update_layout(self):
        if not self.loaded:
            return

        method = self.ui.sortMethod.currentData()
        self.image_mosaic.hide_discarded = False
        self.image_mosaic.hide_to_review = False
        self.image_mosaic.hide_labeled = self.ui.hideLabeled.isChecked()
        self.image_mosaic.render_mosaic(sort_method=method)

    @pyqtSlot(int)
    def update_zoom(self, zoom):
        if not self.loaded:
            return

        self.image_mosaic.update_zoom(zoom / 100)

    @pyqtSlot(object)
    def rect_click(self, rect: RectWidget):
        if not self.loaded:
            return

        # Save information to VARS for any moved/resized boxes
        self.box_handler.save_all(self.verifier)

        # remove highlight from the last selected ROI
        if self.last_selected_roi is not None:
            self.box_handler.clear()
            self.last_selected_roi.isLastSelected = False
            self.last_selected_roi.update()

        # update the new selection
        rect.isLastSelected = True
        rect.update()
        self.last_selected_roi = rect

        full_img = rect.getFullImage()
        if full_img is None:
            return

        # Update the image and add the boxes
        self.box_handler.roiDetail.setImage(cv2.cvtColor(full_img, cv2.COLOR_BGR2RGB))
        self.box_handler.add_annotation(rect.index, rect)

        # Add localization data to the panel
        self.ui.annotationXML.clear()
        self.ui.annotationXML.insertPlainText(
            json.dumps(rect.localization.json, indent=2)
        )

        # Add ancillary data to the image info list
        self.ui.imageInfoList.clear()
        self.ui.imageInfoList.addItem(
            'Derived timestamp: {}'.format(rect.annotation_datetime().strftime('%Y-%m-%d %H:%M:%S'))
        )
        self.ui.imageInfoList.addItems(
            ['{}: {}'.format(key.replace('_', ' ').capitalize(), value) for key, value in rect.ancillary_data.items()]
        )
        
        # Update VARS observations label
        imaged_moment_uuid = rect.localization.imaged_moment_uuid
        if imaged_moment_uuid in self.cached_moment_concepts:  # cache hit
            concepts = self.cached_moment_concepts[imaged_moment_uuid]
        else:  # cache miss
            vars_moment_data = get_vars_imaged_moment(rect.localization.imaged_moment_uuid)
            concepts = sorted(set(obs['concept'] for obs in vars_moment_data['observations']))
            self.cached_moment_concepts[imaged_moment_uuid] = concepts
        observed_concepts_str = ', '.join(concepts)
        self.ui.varsObservationsLabel.setText(observed_concepts_str)
        
    @pyqtSlot()
    def open_video(self):
        """
        Open the video of the last selected ROI, if available
        """
        if not self.last_selected_roi:
            QMessageBox.warning(self, 'No ROI Selected', 'No ROI selected.')
            return
        
        # Get the annotation recorded datetime
        annotation_datetime = self.last_selected_roi.annotation_datetime()
        annotation_platform = self.last_selected_roi.ancillary_data.get('camera_platform', None)
        if not annotation_datetime or not annotation_platform:
            QMessageBox.warning(self, 'Missing Info', 'ROI lacks necessary information to link video.')
            return
        
        # Ask M3 for videos at the given moment
        try:
            videos = get_videos_at_datetime(annotation_datetime)
        except Exception as e:
            QMessageBox.warning(self, 'Error Finding Video', 'An error occurred while finding the video: {}'.format(e))
            return
        
        # Find a matching video URL and timedelta
        video_url = None
        annotation_timedelta = None
        for video in videos:
            if not video['name'].startswith(annotation_platform):  # Skip if platform doesn't match
                continue
            
            video_start_timestamp = video.get('start_timestamp', None)
            if video_start_timestamp is None:  # Skip if no start timestamp in video
                continue
            
            # Parse video start timestamp into datetime object
            try:
                video_start_datetime = datetime.datetime.strptime(video_start_timestamp, '%Y-%m-%dT%H:%M:%S.%fZ')
            except:
                video_start_datetime = datetime.datetime.strptime(video_start_timestamp, '%Y-%m-%dT%H:%M:%SZ')
            
            # Find a matching video reference (MP4 only)
            video_references = video.get('video_references', [])
            for video_reference in video_references:
                if video_reference['uri'].startswith('http') and video_reference['uri'].endswith('.mp4'):
                    video_url = video_reference['uri']
                    annotation_timedelta = annotation_datetime - video_start_datetime
                    break

        # Open the video at the computed time delta (in seconds)
        if video_url is not None:
            annotation_seconds = max(annotation_timedelta.total_seconds(), 0)
            url = video_url + '#t={},{}'.format(annotation_seconds, annotation_seconds + 1e-3)  # "pause" at the annotation
            webbrowser.open(url)
        else:
            QMessageBox.warning(self, 'No Video Found', 'No video found for this ROI.')

    @pyqtSlot()
    def _style_gui(self):
        """
        Set the GUI stylesheet.
        """
        # setup stylesheet
        # set the environment variable to use a specific wrapper
        # it can be set to PyQt, PyQt5, PyQt6 PySide or PySide2 (not implemented yet)
        if self.ui.styleComboBox.currentText().lower() == 'darkstyle':
            os.environ['PYQTGRAPH_QT_LIB'] = 'PyQt6'
            self._app.setStyleSheet(
                qdarkstyle.load_stylesheet(qt_api=os.environ['PYQTGRAPH_QT_LIB'])
            )
        elif self.ui.styleComboBox.currentText().lower() == 'darkbreeze':
            file = QFile(str(ASSETS_DIR / 'style'/ 'dark.qss'))
            file.open(QFile.OpenModeFlag.ReadOnly | QFile.OpenModeFlag.Text)
            stream = QTextStream(file)
            self._app.setStyleSheet(stream.readAll())
        elif self.ui.styleComboBox.currentText().lower() == 'default':
            self._app.setStyleSheet("")

    def closeEvent(self, event):
        self._save_gui()
        if self.loaded:
            self.box_handler.save_all(self.verifier)
        QMainWindow.closeEvent(self, event)

    def query(self) -> Optional[dict]:
        dialog = QueryDialog(parent=self)
        ok = dialog.exec()
        return dialog.constraints_dict() if ok else None


def init_settings():
    """
    Initialize the application settings.
    """
    settings = SettingsManager.get_instance()
    
    settings.sql_url = ('sql/url', str, constants.SQL_URL_DEFAULT)
    settings.sql_user = ('sql/user', str, constants.SQL_USER_DEFAULT)
    settings.sql_password = ('sql/password', str, constants.SQL_PASSWORD_DEFAULT)
    settings.sql_database = ('sql/database', str, constants.SQL_DATABASE_DEFAULT)
    
    settings.raz_url = ('m3/raz_url', str, constants.RAZIEL_URL_DEFAULT)
    
    
def main():
    # Create the Qt application
    app = QApplication(sys.argv)
    app.setApplicationName(constants.APP_NAME)
    app.setOrganizationName(constants.APP_ORGANIZATION)
    
    QSettings.setDefaultFormat(QSettings.Format.IniFormat)
    init_settings()    

    # Create the main window and show it
    main = MainWindow(app)
    main.show()

    # Exit after app is finished
    status = app.exec()
    sys.exit(status)


if __name__ == "__main__":
    main()



from typing import Any, Iterable, List, Optional
from uuid import UUID

from PyQt6.QtCore import QAbstractListModel, QModelIndex, QObject, Qt, pyqtSlot
from PyQt6.QtWidgets import (
    QAbstractItemView,
    QDialog,
    QDialogButtonBox,
    QHBoxLayout,
    QInputDialog,
    QLineEdit,
    QListView,
    QMessageBox,
    QPushButton,
    QStyle,
    QVBoxLayout,
    QWidget,
)

from vars_gridview.lib.m3.operations import get_kb_concepts, get_kb_descendants, get_users


class Constraint:
    def __init__(self, key, value):
        self.key = key
        self.value = value


class Filter:
    class Result:
        @property
        def constraints(self) -> Iterable[Constraint]:
            raise NotImplementedError()

        def __str__(self) -> str:
            raise NotImplementedError()

    def __init__(self, parent, name: str):
        self.parent = parent
        self.name = name

    def __call__(self) -> Optional[Result]:
        raise NotImplementedError()


class ConceptFilter(Filter):
    class Result(Filter.Result):
        def __init__(self, concept: str):
            self.concept = concept

        @property
        def constraints(self) -> Iterable[Constraint]:
            yield Constraint('concept', self.concept)

        def __str__(self) -> str:
            return 'Concept: {}'.format(self.concept)

    def __call__(self) -> Optional[Result]:
        concept, ok = QInputDialog.getItem(
            self.parent, 'Concept', 'Concept', get_kb_concepts(), 0, True
        )
        if ok:
            return ConceptFilter.Result(concept)


class ConceptDescFilter(Filter):
    class Result(ConceptFilter.Result):
        def __init__(self, concept: str):
            super().__init__(concept)
            self.descendants = get_kb_descendants(concept)

        @property
        def constraints(self) -> Iterable[Constraint]:
            yield from super().constraints
            for descendant in self.descendants:
                yield Constraint('concept', descendant)

        def __str__(self) -> str:
            return 'Concept (+ descendants): {} ({})'.format(
                self.concept, ', '.join(self.descendants)
            )

    def __call__(self) -> Optional[Result]:
        concept, ok = QInputDialog.getItem(
            self.parent, 'Concept', 'Concept', get_kb_concepts(), 0, True
        )
        if ok:
            return ConceptDescFilter.Result(concept)


class DiveNumberFilter(Filter):
    class Result(Filter.Result):
        def __init__(self, dive_number: str):
            self.dive_number = dive_number

        @property
        def constraints(self) -> Iterable[Constraint]:
            yield Constraint('dive_number', self.dive_number)

        def __str__(self) -> str:
            return 'Dive number: {}'.format(self.dive_number)

    def __call__(self) -> Optional[Result]:
        dive_number, ok = QInputDialog.getText(
            self.parent, 'Dive number', 'Dive number', QLineEdit.EchoMode.Normal, ''
        )
        if ok:
            return DiveNumberFilter.Result(dive_number)


class ChiefScientistFilter(Filter):
    class Result(Filter.Result):
        def __init__(self, chief_scientist: str):
            self.chief_scientist = chief_scientist

        @property
        def constraints(self) -> Iterable[Constraint]:
            yield Constraint('chief_scientist', self.chief_scientist)

        def __str__(self) -> str:
            return 'Chief scientist: {}'.format(self.chief_scientist)

    def __call__(self) -> Optional[Result]:
        chief_scientist, ok = QInputDialog.getText(
            self.parent, 'Chief scientist', 'Chief scientist', QLineEdit.EchoMode.Normal, ''
        )
        if ok:
            return ChiefScientistFilter.Result(chief_scientist)


class PlatformFilter(Filter):
    class Result(Filter.Result):
        def __init__(self, platform: str):
            self.platform = platform

        @property
        def constraints(self) -> Iterable[Constraint]:
            yield Constraint('platform', self.platform)

        def __str__(self) -> str:
            return 'Platform: {}'.format(self.platform)

    def __call__(self) -> Optional[Result]:
        platform, ok = QInputDialog.getText(
            self.parent, 'Platform', 'Platform', QLineEdit.EchoMode.Normal, ''
        )
        if ok:
            return PlatformFilter.Result(platform)


class ObserverFilter(Filter):
    class Result(Filter.Result):
        def __init__(self, observer: str):
            self.observer = observer

        @property
        def constraints(self) -> Iterable[Constraint]:
            yield Constraint('observer', self.observer)

        def __str__(self) -> str:
            return 'Observer: {}'.format(self.observer)

    def __call__(self) -> Optional[Result]:
        usernames = sorted([user['username'] for user in get_users()])
        observer, ok = QInputDialog.getItem(
            self.parent, 'Observer', 'Observer', usernames, 0, True
        )
        if ok:
            return ObserverFilter.Result(observer)


class ImagedMomentUUIDFilter(Filter):
    class Result(Filter.Result):
        def __init__(self, imaged_moment_uuid: str):
            self.imaged_moment_uuid = imaged_moment_uuid

        @property
        def constraints(self) -> Iterable[Constraint]:
            yield Constraint('imaged_moment_uuid', self.imaged_moment_uuid)

        def __str__(self) -> str:
            return 'Imaged moment UUID: {}'.format(self.imaged_moment_uuid)

    def __call__(self) -> Optional[Result]:
        imaged_moment_uuid, ok = QInputDialog.getText(
            self.parent,
            'Imaged moment UUID',
            'Imaged moment UUID',
            QLineEdit.EchoMode.Normal,
            '',
        )
        if ok:
            # Ensure that the UUID is valid
            try:
                UUID(imaged_moment_uuid)
            except ValueError:
                QMessageBox.warning(self.parent, 'Invalid UUID', 'The UUID is invalid.')
                return None
            return ImagedMomentUUIDFilter.Result(imaged_moment_uuid.lower())


class ObservationUUIDFilter(Filter):
    class Result(Filter.Result):
        def __init__(self, observation_uuid: str):
            self.observation_uuid = observation_uuid

        @property
        def constraints(self) -> Iterable[Constraint]:
            yield Constraint('observation_uuid', self.observation_uuid)

        def __str__(self) -> str:
            return 'Observation UUID: {}'.format(self.observation_uuid)

    def __call__(self) -> Optional[Result]:
        observation_uuid, ok = QInputDialog.getText(
            self.parent, 'Observation UUID', 'Observation UUID', QLineEdit.EchoMode.Normal, ''
        )
        if ok:
            # Ensure that the UUID is valid
            try:
                UUID(observation_uuid)
            except ValueError:
                QMessageBox.warning(self.parent, 'Invalid UUID', 'The UUID is invalid.')
                return None
            return ObservationUUIDFilter.Result(observation_uuid.lower())


class AssociationUUIDFilter(Filter):
    class Result(Filter.Result):
        def __init__(self, association_uuid: str):
            self.association_uuid = association_uuid

        @property
        def constraints(self) -> Iterable[Constraint]:
            yield Constraint('assoc.uuid', self.association_uuid)

        def __str__(self) -> str:
            return 'Association UUID: {}'.format(self.association_uuid)

    def __call__(self) -> Optional[Result]:
        association_uuid, ok = QInputDialog.getText(
            self.parent, 'Association UUID', 'Association UUID', QLineEdit.EchoMode.Normal, ''
        )
        if ok:
            # Ensure that the UUID is valid
            try:
                UUID(association_uuid)
            except ValueError:
                QMessageBox.warning(self.parent, 'Invalid UUID', 'The UUID is invalid.')
                return None
            return AssociationUUIDFilter.Result(association_uuid.lower())


class ImageReferenceUUIDFilter(Filter):
    class Result(Filter.Result):
        def __init__(self, image_reference_uuid: str):
            self.image_reference_uuid = image_reference_uuid

        @property
        def constraints(self) -> Iterable[Constraint]:
            yield Constraint('image_reference_uuid', self.image_reference_uuid)

        def __str__(self) -> str:
            return 'Image reference UUID: {}'.format(self.image_reference_uuid)

    def __call__(self) -> Optional[Result]:
        image_reference_uuid, ok = QInputDialog.getText(
            self.parent,
            'Image reference UUID',
            'Image reference UUID',
            QLineEdit.EchoMode.Normal,
            '',
        )
        if ok:
            # Ensure that the UUID is valid
            try:
                UUID(image_reference_uuid)
            except ValueError:
                QMessageBox.warning(self.parent, 'Invalid UUID', 'The UUID is invalid.')
                return None
            return ImageReferenceUUIDFilter.Result(image_reference_uuid.lower())


class VideoReferenceUUIDFilter(Filter):
    class Result(Filter.Result):
        def __init__(self, video_reference_uuid: str):
            self.video_reference_uuid = video_reference_uuid

        @property
        def constraints(self) -> Iterable[Constraint]:
            yield Constraint('video_reference_uuid', self.video_reference_uuid)

        def __str__(self) -> str:
            return 'Video reference UUID: {}'.format(self.video_reference_uuid)

    def __call__(self) -> Optional[Result]:
        video_reference_uuid, ok = QInputDialog.getText(
            self.parent,
            'Video reference UUID',
            'Video reference UUID',
            QLineEdit.EchoMode.Normal,
            '',
        )
        if ok:
            # Ensure that the UUID is valid
            try:
                UUID(video_reference_uuid)
            except ValueError:
                QMessageBox.warning(self.parent, 'Invalid UUID', 'The UUID is invalid.')
                return None
            return VideoReferenceUUIDFilter.Result(video_reference_uuid.lower())


class ResultListModel(QAbstractListModel):
    def __init__(self, parent: QObject = None, results: List[Filter.Result] = None):
        super().__init__(parent=parent)

        self.results = results or []

    def rowCount(self, parent: QModelIndex = None) -> int:
        return len(self.results)

    def data(self, index: QModelIndex, role: Qt.ItemDataRole) -> Any:
        if not index.isValid():
            return None

        if role == Qt.ItemDataRole.DisplayRole:
            return str(self.results[index.row()])
        else:
            return None

    def add_result(self, result: Filter.Result):
        self.beginInsertRows(QModelIndex(), len(self.results), len(self.results))
        self.results.append(result)
        self.endInsertRows()

    def clear(self):
        self.beginResetModel()
        self.results = []
        self.endResetModel()

    def remove_result(self, index: int):
        self.beginRemoveRows(QModelIndex(), index, index)
        del self.results[index]
        self.endRemoveRows()

    def remove_results(self, indices: List[int]):
        indices.sort(reverse=True)
        for index in indices:
            self.remove_result(index)

    @property
    def constraints(self) -> Iterable[Constraint]:
        for result in self.results:
            yield from result.constraints


class QueryDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent=parent)

        self.setWindowTitle('Query')
        self.setLayout(QVBoxLayout())

        # Create filters
        self.filters = [
            ConceptFilter(self, 'Concept'),
            ConceptDescFilter(self, 'Concept (+ descendants)'),
            DiveNumberFilter(self, 'Dive number'),
            ChiefScientistFilter(self, 'Chief scientist'),
            PlatformFilter(self, 'Platform'),
            ObserverFilter(self, 'Observer'),
            ImagedMomentUUIDFilter(self, 'Imaged moment UUID'),
            ObservationUUIDFilter(self, 'Observation UUID'),
            AssociationUUIDFilter(self, 'Association UUID'),
            ImageReferenceUUIDFilter(self, 'Image reference UUID'),
            VideoReferenceUUIDFilter(self, 'Video reference UUID'),
        ]

        # Create button bar (add, remove, clear constraints)
        self.button_bar = QWidget()
        self.button_bar.setLayout(QHBoxLayout())
        self.add_constraint_button = QPushButton('Add Constraint')
        self.add_constraint_button.setIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_DialogYesButton)
        )
        self.remove_constraint_button = QPushButton('Remove Selected')
        self.remove_constraint_button.setIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_DialogNoButton)
        )
        self.clear_filters_button = QPushButton('Clear')
        self.clear_filters_button.setIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_DialogDiscardButton)
        )

        # Create result list model and view
        self.result_list_model = ResultListModel(parent=self)
        self.result_list_view = QListView()
        self.result_list_view.setModel(self.result_list_model)
        self.result_list_view.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

        # Create dialog button box (just Ok button)
        self.dialog_buttons = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok)

        # Arrange
        self.layout().addWidget(self.button_bar)
        self.layout().addWidget(self.result_list_view)
        self.layout().addWidget(self.dialog_buttons)
        self.button_bar.layout().addWidget(self.add_constraint_button)
        self.button_bar.layout().addWidget(self.remove_constraint_button)
        self.button_bar.layout().addWidget(self.clear_filters_button)

        # Connect signals and slots
        self.dialog_buttons.accepted.connect(self.accept)
        self.add_constraint_button.pressed.connect(self.add_filter)
        self.remove_constraint_button.pressed.connect(self.remove_selected_filters)
        self.clear_filters_button.pressed.connect(self.result_list_model.clear)

    @pyqtSlot()
    def add_filter(self):
        filter_names = [f.name for f in self.filters]
        filter_name, ok = QInputDialog.getItem(
            self, 'Filter', 'Filter', filter_names, 0, False
        )
        if not ok:
            return

        filter = self.filters[filter_names.index(filter_name)]

        result = filter()

        if result is None:
            return

        self.result_list_model.add_result(result)

    @pyqtSlot()
    def remove_selected_filters(self):
        indices = [index.row() for index in self.result_list_view.selectedIndexes()]
        self.result_list_model.remove_results(indices)

    @pyqtSlot()
    def clear(self):
        self.result_list_model.clear()

    def constraints_dict(self):
        d = {}
        for constraint in self.result_list_model.constraints:
            if constraint.key not in d:
                d[constraint.key] = []
            d[constraint.key].append(constraint.value)
        return d



from PyQt6 import QtWidgets

from vars_gridview.ui.settings.tabs.AbstractSettingsTab import AbstractSettingsTab


class M3Tab(AbstractSettingsTab):
    """
    M3 microservices tab.
    """
    
    def __init__(self, parent=None):
        super().__init__('M3', parent=parent)
        
        self.raziel_url_edit = QtWidgets.QLineEdit(self._settings.raz_url.value)
        self.raziel_url_edit.textChanged.connect(self.settingsChanged.emit)
        
        self.arrange()
    
    def arrange(self):
        layout = QtWidgets.QFormLayout()
        
        layout.addRow('Raziel URL', self.raziel_url_edit)
        
        self.setLayout(layout)
    
    def apply_settings(self):
        self._settings.raz_url.value = self.raziel_url_edit.text()



from pathlib import Path

# Application
APP_NAME = 'VARS GridView'
APP_ORGANIZATION = 'MBARI'

# SQL
SQL_URL_DEFAULT = 'perseus.shore.mbari.org'
SQL_USER_DEFAULT = 'everyone'
SQL_PASSWORD_DEFAULT = 'guest'
SQL_DATABASE_DEFAULT = 'M3_ANNOTATIONS'

# M3
RAZIEL_URL_DEFAULT = 'http://m3.shore.mbari.org/config'

# Asset paths
ROOT_DIR = Path(__file__).parent.parent
ASSETS_DIR = ROOT_DIR / 'assets'
UI_FILE = ASSETS_DIR / 'gridview.ui'
STYLE_DIR = ASSETS_DIR / 'style'
GUI_SETTINGS_FILE = ASSETS_DIR / 'gui.ini'
BASE_QUERY_FILE = ASSETS_DIR / 'base_query.sql'
LOG_DIR = ROOT_DIR / 'logs'


from vars_gridview.lib.m3.clients import AnnosaurusClient, VARSKBServerClient, VARSUserServerClient, VampireSquidClient


ANNOSAURUS_CLIENT: AnnosaurusClient = None
VAMPIRE_SQUID_CLIENT: VampireSquidClient = None
VARS_USER_SERVER_CLIENT: VARSUserServerClient = None
VARS_KB_SERVER_CLIENT: VARSKBServerClient = None


def setup_from_endpoint_data(endpoints: list):
    """
    Set up the clients from a list of Raziel endpoint dicts.
    Raise a ValueError if the list does not contain the required information.
    """
    def get_client_url_secret(name: str):
        data = next((e for e in endpoints if e['name'] == name), None)
        if data is None:
            raise ValueError(f'Endpoint "{name}" not found')
        return data['url'], data['secret']
    
    anno_url, anno_api_key = get_client_url_secret('annosaurus')
    global ANNOSAURUS_CLIENT
    ANNOSAURUS_CLIENT = AnnosaurusClient(anno_url)
    ANNOSAURUS_CLIENT.authenticate(anno_api_key)
    
    vam_url, _ = get_client_url_secret('vampire-squid')
    global VAMPIRE_SQUID_CLIENT
    VAMPIRE_SQUID_CLIENT = VampireSquidClient(vam_url)
    
    users_url, _ = get_client_url_secret('vars-user-server')
    global VARS_USER_SERVER_CLIENT
    VARS_USER_SERVER_CLIENT = VARSUserServerClient(users_url)
    
    kb_url, _ = get_client_url_secret('vars-kb-server')
    global VARS_KB_SERVER_CLIENT
    VARS_KB_SERVER_CLIENT = VARSKBServerClient(kb_url)



from base64 import b64encode

import requests


def authenticate(url: str, username: str, password: str):
    """
    Authenticate with Raziel and get the endpoint information.
    """
    # Encode the username and password
    user_pass_base64 = 'Basic ' + b64encode(
        '{}:{}'.format(username, password).encode('utf-8')
    ).decode('utf-8')

    # Attempt to authenticate with Raziel
    res = requests.post(url + '/auth', headers={'Authorization': user_pass_base64})
    res.raise_for_status()

    # Get the token from the response
    token = res.json()['accessToken']

    # Get the endpoints from Raziel
    return requests.get(url + '/endpoints', headers={'Authorization': 'Bearer ' + token}).json()



# -*- coding: utf-8 -*-
"""
image_mosaic.py -- A set of classes to extend widgets from pyqtgraph and pyqt for annotation purposes
Copyright 2020  Monterey Bay Aquarium Research Institute
Distributed under MIT license. See license.txt for more infomation.

"""

from datetime import datetime
from typing import List, Optional
from uuid import UUID

import cv2
import numpy as np
import pyqtgraph as pg
import requests
from PyQt6 import QtWidgets, QtCore

from vars_gridview.lib.annotation import VARSLocalization
from vars_gridview.lib.log import LOGGER
from vars_gridview.lib.sort_methods import SortMethod
from vars_gridview.lib.widgets import RectWidget

IMAGE_TYPE = 'image/png'


class ImageMosaic:
    def __init__(
        self, graphicsView, query_data, query_headers, rect_slot, verifier, beholder_url, beholder_api_key, zoom=1.0
    ):
        self.graphicsView = graphicsView
        self.layout = None
        self.scene = None
        self.panel = None
        self.thumbs = []
        self.roi_map = {}
        self.hide_labeled = True
        self.hide_discarded = True
        self.hide_to_review = True
        self.layouts = []
        # The scene and view for image mosaic
        self.scene = QtWidgets.QGraphicsScene()
        self.graphicsView.setScene(self.scene)
        self.panel = QtWidgets.QGraphicsWidget()
        self.scene.addItem(self.panel)

        self.verifier = verifier

        self.moment_image_data = {}
        self.moment_localizations = {}
        self.moment_video_data = {}
        self.moment_mp4_data = {}
        self.observation_data = {}
        self.moment_image_map = {}  # Big

        self.moment_localizations = {}
        self.moment_ancillary_data = {}
        
        self.beholder_url = beholder_url
        self.beholder_api_key = beholder_api_key

        self.n_images = 0
        self.n_localizations = 0

        # Munge query items into corresponding dicts
        seen_associations = set()
        for query_item in (
            dict(zip(query_headers, i)) for i in query_data
        ):  # TODO Make pagination
            for k, v in query_item.items():  # Stringify UUID objects
                if type(v) == UUID:
                    query_item[k] = str(v)

            # Extract fields
            imaged_moment_uuid = query_item['imaged_moment_uuid']
            image_reference_uuid = query_item['image_reference_uuid']
            observation_uuid = query_item['observation_uuid']
            association_uuid = query_item['association_uuid']
            
            image_format = query_item['image_format']
            image_url = query_item['image_url']
            
            concept = query_item['concept']
            
            link_name = query_item['link_name']
            to_concept = query_item['to_concept']
            link_value = query_item['link_value']
            
            # Set up moment data
            if imaged_moment_uuid not in self.moment_image_data and image_format == IMAGE_TYPE:
                self.moment_image_data[imaged_moment_uuid] = {
                    'image_reference_uuid': image_reference_uuid,
                    'image_url': image_url,
                }

            # Tag in ancillary data
            if imaged_moment_uuid not in self.moment_ancillary_data:
                self.moment_ancillary_data[imaged_moment_uuid] = {
                    k: query_item[k]
                    for k in (  # Ancillary data keys
                        'camera_platform',
                        'dive_number',
                        'depth_meters',
                        'latitude',
                        'longitude',
                        'oxygen_ml_per_l',
                        'pressure_dbar',
                        'salinity',
                        'temperature_celsius',
                        'light_transmission',
                    )
                }
            
            # Extract video data
            video_data = {
                k: query_item[k]
                for k in (  # Video data keys
                    'index_elapsed_time_millis',
                    'index_timecode',
                    'index_recorded_timestamp',
                    'video_start_timestamp',
                    'video_uri',
                    'video_container'
                )
            }
            
            # Tag in video data where appropriate
            if video_data['video_uri'] is not None:  # valid video
                if imaged_moment_uuid not in self.moment_video_data:
                    self.moment_video_data[imaged_moment_uuid] = video_data
                    
                # For beholder
                if imaged_moment_uuid not in self.moment_mp4_data and video_data['video_container'] == 'video/mp4':
                    self.moment_mp4_data[imaged_moment_uuid] = video_data

            # Collect bounding boxes
            if link_name == 'bounding box':
                if association_uuid in seen_associations:
                    continue
                seen_associations.add(association_uuid)

                json_loc = link_value
                localization = VARSLocalization.from_json(json_loc)

                localization.set_concept(concept, to_concept)
                
                localization.imaged_moment_uuid = imaged_moment_uuid
                localization.observation_uuid = observation_uuid
                localization.association_uuid = association_uuid

                if imaged_moment_uuid not in self.moment_localizations:
                    self.moment_localizations[imaged_moment_uuid] = []
                self.moment_localizations[imaged_moment_uuid].append(localization)

        # Create a worklist (imaged moment UUID -> URL or None)
        worklist = {}
        for imaged_moment_uuid in self.moment_localizations:
            if imaged_moment_uuid in self.moment_image_data:
                worklist[imaged_moment_uuid] = self.moment_image_data[imaged_moment_uuid]['image_url']
            else:
                worklist[imaged_moment_uuid] = None

        # Download the images
        with pg.ProgressDialog(
            "Downloading images...", 0, len(set(worklist.keys()))
        ) as dlg:
            for imaged_moment_uuid, url in worklist.items():
                if dlg.wasCanceled():
                    raise Exception("Image loading cancelled by user")
                
                if imaged_moment_uuid in self.moment_image_map:  # already downloaded, skip
                    continue

                if url is None:  # no image reference, need to use beholder
                    elapsed_time_millis = None
                    mp4_data = self.moment_mp4_data.get(imaged_moment_uuid, None)
                    if mp4_data is None:
                        LOGGER.warn('No web video available for capture for moment: {}, skipping'.format(imaged_moment_uuid))
                        continue
                    
                    video_start_datetime = mp4_data['video_start_timestamp']
        
                    elapsed_time_millis = mp4_data.get('index_elapsed_time_millis', None)
                    timecode = mp4_data.get('index_timecode', None)
                    recorded_timestamp = mp4_data.get('index_recorded_timestamp', None)
                    
                    # Get annotation video time index
                    annotation_datetime = None
                    if recorded_timestamp is not None:
                        annotation_datetime = recorded_timestamp
                    elif elapsed_time_millis is not None:
                        annotation_datetime = video_start_datetime + datetime.timedelta(milliseconds=int(elapsed_time_millis))
                    elif timecode is not None:
                        hours, minutes, seconds, frames = map(int, timecode.split(':'))
                        annotation_datetime = video_start_datetime + datetime.timedelta(hours=hours, minutes=minutes, seconds=seconds)
                    else:
                        LOGGER.error('No time index available for moment: {}, skipping'.format(imaged_moment_uuid))
                        continue
                    
                    # Compute the elapsed time in milliseconds
                    elapsed_time_millis = round((annotation_datetime - video_start_datetime).total_seconds() * 1000)
                    res = requests.post(
                        beholder_url + '/capture', 
                        json={
                            'videoUrl': mp4_data['video_uri'],
                            'elapsedTimeMillis': int(elapsed_time_millis),
                        },
                        headers={
                            'X-Api-Key': self.beholder_api_key
                        }
                    )  # TODO REMOVE
                    if res.status_code != 200:
                        LOGGER.error('Error getting capture from beholder for moment: {}, skipping'.format(imaged_moment_uuid))
                        continue
                    
                else:  # download the image from its URL
                    res = requests.get(url)
                    if res.status_code != 200:
                        LOGGER.warn(
                            'Unable to fetch image (status {}) at url: {}, skipping'.format(
                                res.status_code, url
                            )
                        )
                        continue

                img_raw = res.content
                img_arr = np.fromstring(img_raw, np.uint8)
                img = cv2.imdecode(img_arr, cv2.IMREAD_COLOR)

                self.moment_image_map[imaged_moment_uuid] = img
                self.n_images += 1

                dlg += 1

        with pg.ProgressDialog(
            'Creating widgets...', 0, self.n_localizations
        ) as roi_pd:
            for (
                imaged_moment_uuid,
                localizations,
            ) in self.moment_localizations.items():
                # If image not there for one reason or another, skip
                if (
                    imaged_moment_uuid not in self.moment_image_map
                    or self.moment_image_map[imaged_moment_uuid] is None
                ):
                    continue

                image = self.moment_image_map[imaged_moment_uuid]
                ancillary_data = (
                    self.moment_ancillary_data.get(imaged_moment_uuid, None) or {}
                )
                video_data = (
                    self.moment_video_data.get(imaged_moment_uuid, None) or {}
                )
                min_x = 0
                min_y = 0
                max_x = image.shape[1]
                max_y = image.shape[0]

                # Filter out invalid boxes
                localizations = [
                    loc
                    for loc in localizations
                    if loc.valid_box and loc.in_bounds(min_x, min_y, max_x, max_y)
                ]

                # create the widgets
                for idx, localization in enumerate(localizations):
                    other_locs = list(localizations)
                    other_locs.remove(localization)
                    rw = RectWidget(
                        other_locs + [localization],
                        image,
                        ancillary_data,
                        video_data,
                        len(other_locs),
                    )
                    rw.text_label = localization.text_label
                    rw.update_zoom(zoom)
                    rw.rectHover.connect(rect_slot)
                    self.thumbs.append(rw)

                    localization.rect = rw  # Back reference

                    self.n_localizations += 1

                    # update progress bar
                    roi_pd += 1

    def render_mosaic(self, sort_method: Optional[SortMethod] = None):
        """
        Load images + annotations and populate the mosaic
        """
        add_thumbs = False
        if self.layout is None:
            add_thumbs = True
            self.panel.setLayout(None)
            self.layout = QtWidgets.QGraphicsLinearLayout(QtCore.Qt.Orientation.Vertical)
            self.layout.setContentsMargins(50, 100, 50, 50)
            self.panel.setLayout(self.layout)

        if sort_method is not None:
            # Sort the thumbnails
            sort_method.sort(self.thumbs)
            add_thumbs = True

        if add_thumbs:
            # Clear the layouts
            for l in self.layouts:
                while l.count():
                    l.removeItem(l.itemAt(0))
                self.layout.removeItem(l)
            self.layouts = []

        columns = 8
        rows = int(len(self.thumbs) / columns)

        i = 0
        while i < len(self.thumbs):
            row_layout = QtWidgets.QGraphicsLinearLayout(QtCore.Qt.Orientation.Horizontal)
            j = 0
            while j < columns and i < len(self.thumbs):
                hide_thumb = False
                if self.hide_to_review:
                    hide_thumb = hide_thumb | self.thumbs[i].forReview
                if self.hide_discarded:
                    hide_thumb = hide_thumb | self.thumbs[i].toDiscard
                if self.hide_labeled:
                    hide_thumb = hide_thumb | self.thumbs[i].isAnnotated

                hide_thumb |= self.thumbs[i].deleted

                if hide_thumb:
                    self.thumbs[i].hide()
                else:
                    self.thumbs[i].show()

                if add_thumbs:
                    row_layout.addItem(self.thumbs[i])
                    self.thumbs[i].show()
                    j += 1

                i += 1
            if add_thumbs:
                self.layout.addItem(row_layout)
                self.layouts.append(row_layout)

    def apply_label(self, concept, part):
        """
        Apply a label to the selected thumbnails.
        """
        for rect in self.get_selected():
            # Handle empty concept/part
            if concept.strip() == '':  # No concept specified? Verify as-is
                concept = rect.localization.concept
            if part.strip() == '':  # No part specified? ditto
                part = rect.localization.part

            # Set the new concept and immediately push to VARS
            rect.localization.set_verified_concept(concept, part, self.verifier)
            rect.localization.push_changes(self.verifier)

            # Update the widget's text label and deselect it
            rect.text_label = rect.localization.text_label
            rect.isSelected = False

            # Propagate visual changes
            rect.update()

        self.render_mosaic()

    def get_selected(self) -> List[RectWidget]:
        """
        Get selected thumbs (list of RectWidget).
        """
        return [rect for rect in self.thumbs if rect.isSelected]

    def delete_selected(self):
        """
        Delete all selected thumbs and re-render.
        """
        for rect in self.get_selected():
            rect.localization.delete()
            rect.deleted = True

        self.clear_selected()

        self.render_mosaic()

    def clear_selected(self):
        """
        Clear the selection of thumbs.
        """
        for ind in range(0, len(self.thumbs)):
            self.thumbs[ind].isSelected = False
            self.thumbs[ind].update()
        # self.render_mosaic()

    def update_zoom(self, zoom):
        for rect in self.thumbs:
            rect.update_zoom(zoom)



"""
M3 operations. Make use of the clients defined in __init__.py.
"""

from datetime import datetime
import json
from typing import List

import requests

from vars_gridview.lib import m3


KB_CONCEPTS = None
KB_PARTS = None
USERS = None


def get_kb_concepts() -> List[str]:
    """
    Get a list of all concepts in the KB.
    """
    global KB_CONCEPTS
    if not KB_CONCEPTS:    
        response = m3.VARS_KB_SERVER_CLIENT.get_concepts()
        
        response.raise_for_status()
        KB_CONCEPTS = response.json()
    
    return KB_CONCEPTS


def get_kb_parts() -> List[str]:
    """
    Get a list of all parts in the KB.
    """
    global KB_PARTS
    if not KB_PARTS:
        response = m3.VARS_KB_SERVER_CLIENT.get_parts()
        
        response.raise_for_status()
        KB_PARTS = [part['name'] for part in response.json()]
    
    return KB_PARTS


def get_kb_descendants(concept: str) -> List[str]:
    """
    Get a list of all descendants of a concept in the KB, including the concept.
    """
    response = m3.VARS_KB_SERVER_CLIENT.get_phylogeny_taxa(concept)
    
    if response.status_code == 404:
        return []  # concept not found, so no descendants
    else:
        response.raise_for_status()
    
    parsed_response = response.json()
    return [taxa['name'] for taxa in parsed_response]


def get_users() -> List[dict]:
    """
    Get a list of all users as dicts.
    """
    global USERS
    if not USERS:
        response = m3.VARS_USER_SERVER_CLIENT.get_all_users()
        
        response.raise_for_status()
        USERS = response.json()
    
    return USERS


def update_bounding_box_data(association_uuid: str, box_dict: dict) -> dict:
    """
    Update a bounding box's JSON data (link_value field of association).
    """
    request_data = {
        'link_value': json.dumps(box_dict),
    }
    
    response = m3.ANNOSAURUS_CLIENT.update_association(association_uuid, request_data)
    response.raise_for_status()
    return response.json()


def update_bounding_box_part(association_uuid: str, part: str) -> dict:
    """
    Update a bounding box's part (to_concept field of association).
    """
    request_data = {
        'to_concept': part,
    }
    
    response = m3.ANNOSAURUS_CLIENT.update_association(association_uuid, request_data)
    response.raise_for_status()
    return response.json()


def update_observation_concept(observation_uuid: str, concept: str, observer: str) -> dict:
    """
    Update an observation's concept and observer.
    """
    request_data = {
        'concept': concept,
        'observer': observer,  # when we update the concept, we also update the observer
    }
    
    response = m3.ANNOSAURUS_CLIENT.update_observation(observation_uuid, request_data)
    
    response.raise_for_status()
    return response.json()


def delete_bounding_box(association_uuid: str):
    """
    Delete a bounding box.
    """
    response = m3.ANNOSAURUS_CLIENT.delete_association(association_uuid)
    
    response.raise_for_status()


def get_videos_at_datetime(dt: datetime) -> List[dict]:
    """
    Get a list of videos occurring at a given instant.
    """
    timestamp = dt.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
    
    response = m3.VAMPIRE_SQUID_CLIENT.get_videos_at_timestamp(timestamp)
    
    response.raise_for_status()
    return response.json()


def get_vars_imaged_moment(image_reference_uuid: str) -> dict:
    """
    Get MBARI VARS imaged moment by UUID.
    This will not work unless you are connected to the MBARI network!
    """
    response = requests.get('http://m3.shore.mbari.org/anno/v1/imagedmoments/' + image_reference_uuid)
    
    return response.json()




from abc import ABC, abstractmethod
from typing import Any, List

import numpy as np

from vars_gridview.lib.widgets import RectWidget


class SortMethod(ABC):
    """
    Abstract class for named sorting methods of RectWidgets.
    """
    NAME = None
    
    @staticmethod
    @abstractmethod
    def key(rect: RectWidget) -> Any:
        """
        Returns the key used for sorting.
        """
        raise NotImplementedError
    
    @classmethod
    def sort(cls, rect_widgets: List[RectWidget], **kwargs):
        """
        Sort a list of RectWidgets.
        """
        rect_widgets.sort(key=cls.key, **kwargs)


class ImageReferenceUUIDSort(SortMethod):
    NAME = 'Image reference UUID'

    @staticmethod
    def key(rect: RectWidget) -> str:
        return rect.localization.image_reference_uuid or 0


class LabelSort(SortMethod):
    NAME = 'Label'

    @staticmethod
    def key(rect: RectWidget) -> str:
        return rect.text_label


class WidthSort(SortMethod):
    NAME = 'Width'
    
    @staticmethod
    def key(rect: RectWidget) -> int:
        return rect.localization.width


class HeightSort(SortMethod):
    NAME = 'Height'
    
    @staticmethod
    def key(rect: RectWidget) -> int:
        return rect.localization.height


class AreaSort(SortMethod):
    NAME = 'Area'

    @staticmethod
    def key(rect: RectWidget) -> int:
        return WidthSort.key(rect) * HeightSort.key(rect)


class MeanIntensitySort(SortMethod):
    NAME = 'Mean intensity'
    
    @staticmethod
    def key(rect: RectWidget) -> float:
        return np.mean(rect.roi, axis=(0, 1, 2))


class MeanHueSort(SortMethod):
    NAME = 'Mean hue'
    
    @staticmethod
    def key(rect: RectWidget) -> float:
        # RGB -> Hue from https://stackoverflow.com/a/23094494
        means = np.mean(rect.roi, axis=(0, 1))
        means = means / 255
        r, g, b = means
        min_idx, min_val = min(enumerate(means), key=lambda x: x[1])
        max_val = max(means)
        
        if min_idx == 0:
            return (g - b) / (max_val - min_val)
        elif min_idx == 1:
            return 2 + (b - r) / (max_val - min_val)
        else:
            return 4 + (r - g) / (max_val - min_val)


class RegionMeanHueSort(SortMethod):
    NAME = 'Region mean hue (center 1/3)'
    
    @staticmethod
    def key(rect: RectWidget) -> float:
        sub_roi = rect.roi[
            rect.localization.height // 3:rect.localization.height * 2 // 3,
            rect.localization.width // 3:rect.localization.width * 2 // 3
        ]
        
        # RGB -> Hue from https://stackoverflow.com/a/23094494
        means = np.mean(sub_roi, axis=(0, 1))
        means = means / 255
        r, g, b = means
        min_idx, min_val = min(enumerate(means), key=lambda x: x[1])
        max_val = max(means)
        
        if min_idx == 0:
            return (g - b) / (max_val - min_val)
        elif min_idx == 1:
            return 2 + (b - r) / (max_val - min_val)
        else:
            return 4 + (r - g) / (max_val - min_val)


class DepthSort(SortMethod):
    NAME = 'Depth'
    
    @staticmethod
    def key(rect: RectWidget) -> float:
        depth = rect.ancillary_data.get('depth_meters', 0.0)
        
        if depth is None:
            depth = 0.0
        
        return depth



# -*- coding: utf-8 -*-
"""
widgets.py -- A set of classes to extend widgets from pyqtgraph and pyqt for annotation purposes
Copyright 2020  Monterey Bay Aquarium Research Institute
Distributed under MIT license. See license.txt for more information.

"""

from typing import List, Optional
import datetime

import cv2
import numpy as np
from PyQt6 import QtCore, QtGui, QtWidgets

from vars_gridview.lib.annotation import VARSLocalization


class RectWidget(QtWidgets.QGraphicsWidget):
    rectHover = QtCore.pyqtSignal(object)

    def __init__(
        self,
        localizations: List[VARSLocalization],
        image: np.ndarray,
        ancillary_data: dict,
        video_data: dict,
        index: int,
        parent=None,
        text_label='rect widget',
    ):
        QtWidgets.QGraphicsWidget.__init__(self, parent)

        self.localizations = localizations  # Dumb, but it works
        self.image = image
        self.ancillary_data = ancillary_data
        self.video_data = video_data
        self.index = index

        self.labelheight = 30
        self.bordersize = 4
        self.picdims = [240, 240]
        self.zoom = 0.5
        self.text_label = text_label
        self._boundingRect = QtCore.QRect()
        self.setAcceptHoverEvents(True)
        self.bgColor = QtCore.Qt.GlobalColor.darkGray
        self.hoverColor = QtCore.Qt.GlobalColor.lightGray
        self.isLastSelected = False
        self.isSelected = False
        self.forReview = False
        self.toDiscard = False

        self.roi = None
        self.pic = None
        self.update_roi_pic()

        self.deleted = False

    def update_roi_pic(self):
        self.roi = self.localization.get_roi(self.image)
        self.pic = self.getpic(self.roi)
        self.update()

    @property
    def isAnnotated(self) -> bool:
        return self.localizations[self.index].verified

    @property
    def localization(self):
        return self.localizations[self.index]

    @property
    def image_width(self):
        return self.image.shape[1]

    @property
    def image_height(self):
        return self.image.shape[0]

    def annotation_datetime(self) -> Optional[datetime.datetime]:
        video_start_datetime = self.video_data['video_start_timestamp']
        
        elapsed_time_millis = self.video_data.get('index_elapsed_time_millis', None)
        timecode = self.video_data.get('index_timecode', None)
        recorded_timestamp = self.video_data.get('index_recorded_timestamp', None)
        
        # Get annotation video time index
        annotation_datetime = None
        if recorded_timestamp is not None:
            annotation_datetime = recorded_timestamp
        elif elapsed_time_millis is not None:
            annotation_datetime = video_start_datetime + datetime.timedelta(milliseconds=int(elapsed_time_millis))
        elif timecode is not None:
            hours, minutes, seconds, frames = map(int, timecode.split(':'))
            annotation_datetime = video_start_datetime + datetime.timedelta(hours=hours, minutes=minutes, seconds=seconds)
        else:
            raise ValueError('No video time index found')
        
        return annotation_datetime

    def toqimage(self, img):
        height, width, bytesPerComponent = img.shape
        bytesPerLine = bytesPerComponent * width
        cv2.cvtColor(img, cv2.COLOR_BGR2RGB, img)
        qimg = QtGui.QImage(
            img.copy(), width, height, bytesPerLine, QtGui.QImage.Format.Format_RGB888
        )

        return qimg

    def update_zoom(self, zoom):
        self.zoom = zoom
        self.boundingRect()
        self.updateGeometry()

    def getFullImage(self):
        return np.rot90(self.image, 3, (0, 1))

    def boundingRect(self):
        # scale and zoom
        width = self.zoom * (self.picdims[0] + self.bordersize * 2)
        height = self.zoom * (self.picdims[1] + self.labelheight + self.bordersize * 2)

        thumb_widget_rect = QtCore.QRectF(0.0, 0.0, width, height)
        self._boundingRect = thumb_widget_rect

        return thumb_widget_rect

    def sizeHint(self, which, constraint=QtCore.QSizeF()):
        return self._boundingRect.size()

    def getpic(self, roi):
        height, width, channels = roi.shape
        if height >= width:
            scale = self.picdims[0] / height
        else:
            scale = self.picdims[0] / width
        new_width = int(width * scale) - 2 * self.bordersize
        new_height = int(height * scale) - 2 * self.bordersize
        roi = cv2.resize(roi, (new_width, new_height))

        # center roi on dims
        w_pad = int((self.picdims[0] - new_width) / 2)
        h_pad = int((self.picdims[1] - new_height) / 2)

        roi = cv2.copyMakeBorder(
            roi,
            h_pad,
            h_pad,
            w_pad,
            w_pad,
            cv2.BORDER_CONSTANT,
            value=(45, 35, 25),
        )

        qimg = self.toqimage(roi)
        orpixmap = QtGui.QPixmap.fromImage(qimg)
        return orpixmap

    def paint(self, painter, option, widget):
        pen = QtGui.QPen()
        pen.setWidth(1)
        pen.setBrush(QtCore.Qt.GlobalColor.black)
        painter.setPen(pen)

        # very simple selection and annotation logic
        if self.isSelected:
            fill_color = QtCore.Qt.GlobalColor.green
        elif self.isAnnotated:
            fill_color = QtCore.Qt.GlobalColor.yellow
        else:
            fill_color = QtCore.Qt.GlobalColor.darkGray

        # fill behind image
        if self.isLastSelected:
            painter.fillRect(
                QtCore.QRect(
                    0,
                    0,
                    int(self.zoom * (self.pic.rect().width() + 2 * self.bordersize)),
                    int(self.zoom)
                    * (
                        self.pic.rect().height()
                        + self.labelheight
                        + 2 * self.bordersize
                    ),
                ),
                QtGui.QColor(61, 174, 233, 255),
            )

        # Fill label
        painter.fillRect(
            QtCore.QRect(
                int(self.zoom * self.bordersize),
                int(self.zoom * (self.bordersize + self.pic.rect().height())),
                int(self.zoom * self.pic.rect().width()),
                int(self.zoom * self.labelheight),
            ),
            fill_color,
        )

        # Draw image
        painter.drawPixmap(
            QtCore.QRect(
                int(self.zoom * self.bordersize),
                int(self.zoom * self.bordersize),
                int(self.zoom * self.pic.rect().width()),
                int(self.zoom * self.pic.rect().height()),
            ),
            self.pic,
            self.pic.rect(),
        )

        # Draw text
        text_rect = QtCore.QRect(
            0,
            int(self.zoom * (self.pic.rect().y() + self.pic.rect().height())),
            int(self.zoom * self.pic.rect().width()),
            int(self.zoom * self.labelheight),
        )

        painter.drawText(text_rect, QtCore.Qt.AlignmentFlag.AlignCenter, self.text_label)

        if self.toDiscard:
            painter.fillRect(
                QtCore.QRect(
                    int(self.zoom * self.bordersize),
                    int(self.zoom * (self.bordersize)),
                    int(self.zoom * self.pic.rect().width()),
                    int(self.zoom * self.labelheight),
                ),
                QtCore.Qt.GlobalColor.gray,
            )
            text_rect = QtCore.QRect(
                0,
                int(self.zoom * (self.pic.rect().y())),
                int(self.zoom * self.pic.rect().width()),
                int(self.zoom * self.labelheight),
            )
            painter.setPen(QtCore.Qt.GlobalColor.red)
            painter.drawText(text_rect, QtCore.Qt.AlignmentFlag.AlignCenter, "To Remove")

        if self.forReview:
            painter.fillRect(
                QtCore.QRect(
                    int(self.zoom * self.bordersize),
                    int(self.zoom * (self.bordersize)),
                    int(self.zoom * self.pic.rect().width()),
                    int(self.zoom * self.labelheight),
                ),
                QtCore.Qt.GlobalColor.gray,
            )
            text_rect = QtCore.QRect(
                0,
                int(self.zoom * (self.pic.rect().y())),
                int(self.zoom * self.pic.rect().width()),
                int(self.zoom * self.labelheight),
            )
            painter.setPen(QtCore.Qt.GlobalColor.blue)
            painter.drawText(text_rect, QtCore.Qt.AlignmentFlag.AlignCenter, "For Review")

    def mousePressEvent(self, event):
        self.isSelected = not self.isSelected
        self.update()
        self.rectHover.emit(self)

    def mouseReleaseEvent(self, event):
        pass

    def hoverEnterEvent(self, event):
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if modifiers == QtCore.Qt.KeyboardModifier.ControlModifier:
            self.isSelected = not self.isSelected
            self.update()
            self.rectHover.emit(self)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vars_gridview',
 'vars_gridview.lib',
 'vars_gridview.lib.m3',
 'vars_gridview.scripts',
 'vars_gridview.ui',
 'vars_gridview.ui.settings',
 'vars_gridview.ui.settings.tabs']

package_data = \
{'': ['*'],
 'vars_gridview': ['assets/base_query.sql',
                   'assets/base_query.sql',
                   'assets/gridview.ui',
                   'assets/gridview.ui',
                   'assets/style/*']}

install_requires = \
['opencv-python-headless>=4.5.5.62,<5.0.0.0',
 'pymssql>=2.2.4,<3.0.0',
 'pyqt6>=6.4.0,<7.0.0',
 'pyqtgraph>=0.13.0,<0.14.0',
 'qdarkstyle>=3.0.3,<4.0.0',
 'requests>=2.27.1,<3.0.0']

entry_points = \
{'console_scripts': ['vars-gridview = vars_gridview.scripts.run:main']}

setup_kwargs = {
    'name': 'vars-gridview',
    'version': '0.1.0',
    'description': 'VARS GridView is a tool for reviewing and correcting VARS localizations in bulk.',
    'long_description': '# vars-gridview\n\n**VARS GridView** is a tool for reviewing and correcting VARS localizations in bulk.\n\n[![MIT License](https://img.shields.io/badge/license-MIT-blue.svg)](https://opensource.org/licenses/MIT)\n[![Python](https://img.shields.io/badge/language-Python-blue.svg)](https://www.python.org/downloads/)\n\nAuthors: Kevin Barnard ([kbarnard@mbari.org](mailto:kbarnard@mbari.org)), Paul Roberts ([proberts@mbari.org](mailto:proberts@mbari.org))\n\n---\n\n## Install\n\n### From PyPI\n\nVARS GridView is available on PyPI as `vars-gridview`. To install, run:\n\n```bash\npip install vars-gridview\n```\n\n### From source\n\nThis project is built with [Poetry](https://python-poetry.org/). To install from source, run (in the project root):\n\n```bash\npoetry install\n```\n\n## Run\n\nOnce VARS GridView is installed, you can run it from the command line:\n\n```bash\nvars-gridview\n```\n\nYou will first be prompted to log in. Enter your VARS username and password. \n\n*Note: If you are not using MBARI production VARS, change the "Config server" field to point to your instance of Raziel. This setting is persisted.*\n\n---\n\nCopyright &copy; 2020&ndash;2022 [Monterey Bay Aquarium Research Institute](https://www.mbari.org)',
    'author': 'Kevin Barnard',
    'author_email': 'kbarnard@mbari.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mbari-org/vars-gridview',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""
rectlabel.py -- Tools to implement a labeling UI for bounding boxes in images
Copyright 2020  Monterey Bay Aquarium Research Institute
Distributed under MIT license. See license.txt for more information.

"""

import numpy as np
import pyqtgraph as pg
from PyQt6 import QtCore

from vars_gridview.lib.log import LOGGER


class BoundingBox(pg.RectROI):
    def __init__(
        self,
        view,
        pos,
        size,
        rect,
        localization,
        verifier,
        color=(255, 0, 0),
        label='ROI',
    ):
        pg.RectROI.__init__(
            self,
            pos,
            size,
            pen=pg.mkPen(color, width=3, style=QtCore.Qt.PenStyle.DashLine),
            invertible=True,
            rotatable=False,
            removable=False,
            sideScalers=True,
        )
        # GUI things
        self.addScaleHandle([1, 0], [0, 1])
        self.addScaleHandle([0, 1], [1, 0])
        self.addTranslateHandle([0.5, 0.5])
        self.sigRemoveRequested.connect(self.remove)
        self.sigRegionChanged.connect(self.draw_name)

        self.sigRegionChangeFinished.connect(self.check_bounds)

        self.label = label
        self.color = color
        self.textItem = None

        self.view = view
        self.view.addItem(self)
        self.draw_name()

        self.verifier = verifier

        self.dirty = False
        self.sigRegionChanged.connect(self._dirty)

        self.sigRegionChangeFinished.connect(self.update_localization_box)

        self.localization = localization
        self.rect = rect

    def check_bounds(self):
        x, y = self.pos()
        w, h = self.size()
        h = -h  # Fix negative height

        min_x = 0
        max_x = self.rect.image_width - w

        min_y = h
        max_y = self.rect.image_height

        if x < min_x or y < min_y or x > max_x or y > max_y:
            self.setPos(min(max_x, max(min_x, x)), min(max_y, max(min_y, y)))

    @property
    def isSelected(self):
        return self.rect.isSelected

    def _dirty(self):
        self.dirty = True

    def update_localization_box(self):
        x, y, w, h = self.get_box()
        y = self.rect.image_height - y
        self.localization.set_box(x, y, w, h)
        self.localization.rect.update_roi_pic()

    def update_label(self):
        self.label = self.localization.text_label
        self.draw_name()

    def remove(self, dummy):
        self.view.removeItem(self.textItem)
        self.view.removeItem(self)

    def get_box(self):
        x, y = self.pos()
        w, h = self.size()

        # need to convert box coordinates to account for positive or
        # negative width/height values from pyqtgraph roi object
        if h > 0:
            y = y + h
        if w < 0:
            x = x + w

        return [x, y, np.abs(w), np.abs(h)]

    def draw_name(self):
        roiSize = self.size()
        w = roiSize[0]
        h = roiSize[1]
        if w < 0:
            x = self.pos().x() - np.abs(w)
        else:
            x = self.pos().x()
        if h < 0:
            y = self.pos().y() - np.abs(h)
        else:
            y = self.pos().y()

        if self.textItem is None:

            self.textItem = pg.TextItem(
                text=self.label, color=np.array(self.color) / 2, fill=(200, 200, 200)
            )
            self.textItem.setPos(x, y)
            self.view.addItem(self.textItem)
        else:
            self.textItem.setText(self.label)
            self.textItem.setPos(x, y)

    def mouseClickEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.RightButton:  # Capture and ignore right clicks
            pass


class BoxHandler:
    def __init__(
        self,
        graphics_view,
        image_mosaic,
        localization=None,
        all_labels=[],
        verifier=None,
    ):
        self.boxes = []
        self.dragging = False

        self.vb = pg.ViewBox()
        self.graphics_view = graphics_view
        self.graphics_view.setCentralItem(self.vb)
        self.vb.setAspectLocked()
        self.roiDetail = pg.ImageItem()
        self.vb.addItem(self.roiDetail)
        self.localization = localization
        self.all_labels = all_labels
        self.verifier = verifier

        self.image_mosaic = image_mosaic

    def update_labels(self):
        for box in self.boxes:
            if box.isSelected:
                box.update_label()

    def add_annotation(self, obj_idx, rect):
        for idx, localization in enumerate(rect.localizations):
            if localization.deleted:
                continue
            selected_loc = idx == obj_idx

            # Set color of the box
            color = (52, 161, 235)  # Nice light blue
            if not selected_loc:
                avg = int(sum(color) / len(color))
                color = (avg, avg, avg)  # Grayscale equivalent of original color

            # Grab the bounds
            xmin, ymin, xmax, ymax = localization.box
            if xmax - xmin <= 0 or ymax - ymin <= 0:  # Bad box bounds check
                LOGGER.warn('Bad box bounds, not adding to the view')
            else:
                label = localization.text_label
                height = rect.image_height

                # Create the bounding box
                bb = BoundingBox(
                    self.vb,
                    [xmin, height - ymin],
                    [xmax - xmin, -1 * (ymax - ymin)],
                    rect,
                    localization,
                    self.verifier,
                    color=color,
                    label=label,
                )

                # Add it to the list
                self.boxes.append(bb)

        self.localization = rect.localizations[obj_idx]

    def save_all(self, verifier):
        for box in self.boxes:
            if box.dirty:
                box.localization.push_changes(verifier)

    def mapToItem(self, pos):
        pt = self.vb.mapSceneToView(pos)
        pt = self.vb.mapFromViewToItem(self.roiDetail, pt)
        return pt

    def clear(self):
        for box in self.boxes:
            box.remove(None)
        self.boxes = []



# -*- coding: utf-8 -*-
"""
rectlabel.py -- Tools to implement a labeling UI for bounding boxes in images
Copyright 2020  Monterey Bay Aquarium Research Institute
Distributed under MIT license. See license.txt for more information.

"""

import logging
from datetime import datetime

from vars_gridview.lib.constants import LOG_DIR


def get_logger(
    filepath, file_level=logging.DEBUG, console_level=logging.WARN, logger_name=''
):
    """Get a logger object initialized with console and file output

    Parameters:
    -----------
    filepath : str
        The absolue path to the file for the FileHandler
    log_level : {'DEBUG', 'INFO', 'WARNING','ERROR','CRITICAL'}
        The log level for the logger,
    logger_name : str, optional
        A string prepended to each log message

    Returns:
    --------
    lgr: logger
        An initialized logger object

    """
    # create logger
    lgr = logging.getLogger(logger_name)
    lgr.setLevel(logging.DEBUG)

    # create console handler and set level to debug
    ch = logging.StreamHandler()
    ch.setLevel(console_level)

    fh = logging.FileHandler(filepath)
    fh.setLevel(file_level)

    # create formatter
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )

    # add formatter to ch
    ch.setFormatter(formatter)
    fh.setFormatter(formatter)

    # add ch to lgr
    lgr.addHandler(ch)
    lgr.addHandler(fh)

    return lgr


# Create a logs dir if it does not exist
LOG_DIR.mkdir(parents=True, exist_ok=True)

LOGGER = get_logger(
    str(LOG_DIR / (datetime.now().strftime('%Y-%m-%d') + '.txt')),
    file_level=logging.DEBUG,
    console_level=logging.WARN,
    logger_name='vars-gridview',
)
